"""
count step
"""

import os
import random
import sys

import numpy as np
import pandas as pd

from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.tools.__init__ import (FILTERED_MATRIX_DIR_SUFFIX, RAW_MATRIX_DIR_SUFFIX)
from celescope.tools.emptydrop_cr import get_plot_elements
from celescope.tools.emptydrop_cr.cell_calling_3 import cell_calling_3
from celescope.tools.step import Step, s_common
from celescope.rna.mkref import Mkref_rna
from celescope.tools.matrix import CountMatrix, ROW, COLUMN
from celescope.tools import reference

TOOLS_DIR = os.path.dirname(__file__)
random.seed(0)
np.random.seed(0)

# downsample.csv
READ_FRACTION = 'read_fraction'
MEDIAN_GENE_NUMBER = 'median_gene_number'
SATURATION = 'saturation'

# Plot axis title in HTML
X_TITLE = 'Read Fraction'
SATURATION_Y_TITLE = 'Sequencing Saturation(%)'
MEDIAN_GENE_Y_TITLE = 'Median Genes per Cell'


class Count(Step):
    """
    ## Features
    - Cell-calling: Distinguish cell barcodes from background barcodes. 
    - Generate expression matrix.
    ## Output
    - `{sample}_raw_feature_bc_matrix` The expression matrix of all detected barcodes in [Matrix Market Exchange Formats](
        https://math.nist.gov/MatrixMarket/formats.html). 
    - `{sample}_filtered_feature_bc_matrix` The expression matrix of cell barcodes in Matrix Market Exchange Formats. 
    - `{sample}_count_detail.txt.gz` 4 columns: 
        - barcode  
        - gene ID  
        - UMI count  
        - read_count  
    - `{sample}_counts.txt` 6 columns:
        - Barcode: barcode sequence
        - read: read count of each barcode
        - UMI: UMI count for each barcode
        - geneID: gene count for each barcode
        - mark: cell barcode or backgound barcode.
            `CB` cell  
            `UB` background  
    - `{sample}_downsample.tsv` Subset a fraction of reads and calculate median gene number and sequencing saturation.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.force_cell_num = args.force_cell_num
        self.cell_calling_method = args.cell_calling_method
        self.expected_cell_num = int(args.expected_cell_num)

        # set
        gtf_file = Mkref_rna.get_config(args.genomeDir)['files']['gtf']
        gp = reference.GtfParser(gtf_file)
        gp.get_id_name()
        self.features = gp.get_features()
        self.downsample_dict = {}

        # output files
        self.marked_count_file = f'{self.outdir}/{self.sample}_counts.txt'
        self.raw_matrix_dir = f'{self.outdir}/{self.sample}_{RAW_MATRIX_DIR_SUFFIX}'
        self.cell_matrix_dir = f'{self.outdir}/{self.sample}_{FILTERED_MATRIX_DIR_SUFFIX}'
        self.downsample_file = f'{self.outdir}/{self.sample}_downsample.tsv'

    def get_df_line(self):
        df_line = pd.read_csv(self.downsample_file, sep="\t", header=0)
        df_line.rename(columns={
            READ_FRACTION: X_TITLE,
            MEDIAN_GENE_NUMBER: MEDIAN_GENE_Y_TITLE,
            SATURATION: SATURATION_Y_TITLE,
        }, inplace=True)

        return df_line

    @staticmethod
    def get_df_cell(df, cell_bc):
        """
        reset index in a must. otherwise the index remains the same with the raw matrix
        """
        df_cell = df.loc[df.index.isin(cell_bc, level=0)]
        df_cell = df_cell.reset_index().set_index([COLUMN, ROW])
        return df_cell

    @utils.add_log
    def run(self):
        # index use less than >50% mem than int index
        df = pd.read_table(self.args.count_detail, header=0, dtype={'geneID':str}, index_col=[0,1])
        total_reads = sum(df['read'])
        self.write_sparse_matrix(df, self.raw_matrix_dir)
        df_bc = self.get_df_bc(df)
        # call cells
        cell_bc, _threshold = self.cell_calling(df_bc)
        # write marked_df_sum
        self.write_marked_df_bc(df_bc, cell_bc)
        # export cell matrix
        df_cell = self.get_df_cell(df, cell_bc)
        median_gene = self.get_median_gene(df_cell)
        saturation = self.get_saturation(df_cell)
        del df
        self.write_sparse_matrix(df_cell, self.cell_matrix_dir)
        total_cell_gene = len(df_cell.index.levels[1])
        df_bc_cell = self.get_df_bc(df_cell)
        del df_bc
        # metrics
        self.add_count_metrics(total_reads, total_cell_gene, cell_bc, df_bc_cell, median_gene, saturation)
        # plot
        self.add_plot_data()

    def add_count_metrics(self, total_reads, total_cell_gene, cell_bc, df_sum_cell, median_gene, saturation):
        n_cells = len(cell_bc)
        self.add_metric(
            name='Estimated Number of Cells',
            value=n_cells,
            help_info=f'the number of barcodes considered as cell-associated. The cell calling method used for this sample is {self.cell_calling_method}.'
        )

        cell_reads = sum(df_sum_cell['read'])
        fraction_reads_in_cells = round(float(cell_reads) / total_reads * 100, 2)
        self.add_metric(
            name='Fraction Reads in Cells',
            value=fraction_reads_in_cells,
            display=f'{fraction_reads_in_cells}%',
            help_info='the fraction of uniquely-mapped-to-transcriptome reads with cell-associated barcodes'
        )

        try:
            valid_read_number = self.get_slot_key(
                slot='metrics',
                step_name='barcode',
                key='Valid Reads',
            )
        except KeyError:
            sys.stderr.write('WARNING: Will not output `Mean Reads per Cell`')
        else:
            mean_reads_per_cell = int(valid_read_number / n_cells)
            self.add_metric(
                name='Mean Reads per Cell',
                value=mean_reads_per_cell,
                help_info='the number of valid reads divided by the estimated number of cells'
            )

        median_umi_per_cell = int(np.median(df_sum_cell['UMI']))
        self.add_metric(
            name='Median UMI per Cell',
            value=median_umi_per_cell,
            help_info='the median number of UMI counts per cell-associated barcode'
        )

        self.add_metric(
            name='Total Genes',
            value=total_cell_gene,
            help_info='the number of genes with at least one UMI count in any cell'
        )

        self.add_metric(
            name='Median Genes per Cell',
            value=int(median_gene),
            help_info='the median number of genes detected per cell-associated barcode'
        )

        self.add_metric(
            name='Saturation',
            value=saturation,
            display=f'{saturation}%',
            help_info='the fraction of read originating from an already-observed UMI. '
        )

    def add_plot_data(self):
        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.marked_count_file))


    
    @utils.add_log
    def cell_calling(self, df_sum):
        """
        Returns:
            cell_bc: set
            UMI_threshold: int
        """
        cell_calling_method = self.cell_calling_method

        if (self.force_cell_num is not None) and (self.force_cell_num != 'None'):
            cell_bc, UMI_threshold = self.force_cell(df_sum)
        elif cell_calling_method == 'auto':
            cell_bc, UMI_threshold = self.auto_cell(df_sum)
        elif cell_calling_method == 'EmptyDrops_CR':
            cell_bc, UMI_threshold = self.emptydrop_cr_cell(df_sum)
        cell_bc = set(cell_bc)
        return cell_bc, UMI_threshold

    @utils.add_log
    def force_cell(self, df_sum):
        force_cell_num = int(self.force_cell_num)

        df_barcode_count = df_sum.groupby(
            ['UMI']).size().reset_index(
            name='barcode_counts')
        sorted_df = df_barcode_count.sort_values("UMI", ascending=False)
        sorted_df["barcode_cumsum"] = sorted_df["barcode_counts"].cumsum()
        num_points = sorted_df.shape[0]
        for i in range(num_points):
            if sorted_df.iloc[i, :]["barcode_cumsum"] >= force_cell_num:
                index_low = i - 1
                index_high = i
                break
        df_sub = sorted_df.iloc[index_low: index_high + 1, :]
        distance = abs(df_sub["barcode_cumsum"] - force_cell_num)
        actual_index = np.argmin(distance)
        threshold = df_sub.iloc[actual_index, :]['UMI']
        cell_bc = Count.get_cell_bc(df_sum, threshold, col='UMI')
        return cell_bc, threshold

    @staticmethod
    def find_threshold(df_sum, idx):
        return int(df_sum.iloc[idx - 1, df_sum.columns == 'UMI'])

    @staticmethod
    def get_cell_bc(df_sum, threshold, col='UMI'):
        return list(df_sum[df_sum[col] >= threshold].index)

    @utils.add_log
    def auto_cell(self, df_sum):
        idx = int(self.expected_cell_num * 0.01)
        barcode_number = df_sum.shape[0]
        idx = int(min(barcode_number, idx))
        if idx == 0:
            sys.exit("cell number equals zero!")
        # calculate read counts threshold
        threshold = int(Count.find_threshold(df_sum, idx) * 0.1)
        threshold = max(1, threshold)
        cell_bc = Count.get_cell_bc(df_sum, threshold)

        return cell_bc, threshold

    @utils.add_log
    def emptydrop_cr_cell(self, df_sum):
        cell_bc, initial_cell_num = cell_calling_3(self.raw_matrix_dir, self.expected_cell_num)
        threshold = Count.find_threshold(df_sum, initial_cell_num)
        return cell_bc, threshold

    @staticmethod
    def get_df_bc(df, sort_by='UMI'):
        '''
        Returns:
            df with 3 cols sort by UMI(default)
                read: int
                UMI: int
                geneID: str
        '''

        df_bc = df.groupby('Barcode').agg({
            'UMI': 'sum',
            'read': 'sum',
        })
        df_bc[ROW] = df.groupby(level=[0]).size()
        return df_bc.sort_values(sort_by, ascending=False)

    def write_marked_df_bc(self, df_sum, cell_bc):
        df_sum.loc[:, 'mark'] = 'UB'
        df_sum.loc[df_sum.index.isin(cell_bc), 'mark'] = 'CB'
        df_sum.to_csv(self.marked_count_file, sep='\t')

    @utils.add_log
    def write_sparse_matrix(self, df, matrix_dir):

        count_matrix = CountMatrix.from_dataframe(df, self.features, value="UMI")
        count_matrix.to_matrix_dir(matrix_dir)

    @utils.add_log
    def cell_summary(self, df, cell_bc):

        df.loc[:, 'mark'] = 'UB'
        df.loc[df['Barcode'].isin(cell_bc), 'mark'] = 'CB'
        CB_total_Genes = df.loc[df['mark'] == 'CB', 'geneID'].nunique()
        CB_reads_count = df.loc[df['mark'] == 'CB', 'count'].sum()
        reads_mapped_to_transcriptome = df['count'].sum()
        return(CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome)

    @staticmethod
    def get_saturation(df_cell):
        unique = sum(df_cell['unique'])
        dup = sum(df_cell['PCR_duplicate'])
        saturation = (1 - unique / (unique + dup)) * 100
        return round(saturation, 2)

    @staticmethod
    def get_median_gene(df_cell):
        return df_cell.groupby('Barcode')['UMI'].count().median()


@utils.add_log
def count(args):
    with Count(args, display_title="Cells") as runner:
        runner.run()


def get_opts_count(parser, sub_program):
    parser.add_argument('--genomeDir', help='Required. Genome directory.')
    parser.add_argument('--expected_cell_num', help='Expected cell number.', default=8000)
    parser.add_argument(
        '--cell_calling_method',
        help=HELP_DICT['cell_calling_method'],
        choices=['auto', 'EmptyDrops_CR'],
        default='EmptyDrops_CR',
    )
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--count_detail', help='Required. File from featureCounts.', required=True)
        parser.add_argument(
            '--force_cell_num',
            help='Default `None`. Force the cell number to be this number. ',
        )



from celescope.bulk_rna.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_bulk_rna(Multi):
    """
    ## Usage
    - For default 384 wells:
    ```
        multi_bulk_rna\\
        --mapfile ./bulk_rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus
    ```
    - For 96 wells, set `--wells`:
    ```
        multi_bulk_rna\\
        --mapfile ./bulk_rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --wells 96
    ```

    Work for AccuraCode bulk RNA-Seq.
    """

    def count(self, sample):
        step = 'count'
        count_detail = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_count_detail.txt'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--count_detail {count_detail} '
        )
        self.process_cmd(cmd, step, sample, m=20, x=1)


def main():
    multi = Multi_bulk_rna(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



from celescope.tools import consensus as supert_cs
from celescope.tools import utils
from xopen import xopen
import pysam


class Consensus(supert_cs.Consensus):
    """
    ## Features
    - Consensus all the reads of the same (barcode, UMI) combinations into one read(UMI). It will go through the sequence residue by residue and 
    count up the number of each type of residue (ie. A or G or T or C for DNA) in all sequences in the
    alignment. If the following conditions are met, the consensus sequence will be the most common residue in the alignment:
    1. the percentage of the most common residue type > threshold(default: 0.5);
    2. most common residue reads >= min_consensus_read;
    otherwise an ambiguous character(N) will be added.

    ## Output
    - `{sample}_consensus.fq` Fastq file after consensus.
    """
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        
        self.filtered_consensus_out = f'{self.out_prefix}_filtered_consensus.fasta'
        
    def __exit__(self, *args, **kwargs):
        self.filter_consensus_out()
        self._clean_up()
        
    @utils.add_log
    def filter_consensus_out(self):
        out_h = xopen(self.filtered_consensus_out, 'w')
        n_umi_count = 0
        n_filter_umi_count = 0
        
        with pysam.FastxFile(self.consensus_out) as f:
            for read in f:
                n_umi_count += 1
                #if set(read.sequence) == {'N'}:
                #if read.sequence.count('N') > 10:
                if read.sequence.count('N') > 5:
                    n_filter_umi_count += 1
                else:
                    out_h.write(utils.fasta_line(read.name, read.sequence))
        out_h.close()

        self.add_metric(
            name="Filtered UMI Counts",
            value=n_umi_count - n_filter_umi_count,
            total=n_umi_count,
            help_info="Filtered UMI from Consensus UMI fasta"
        )


@utils.add_log
def consensus(args):
    if args.not_consensus:
        return
    with Consensus(args, display_title="Consensus") as runner:
        runner.run()


def get_opts_consensus(parser, sub_program):
    supert_cs.get_opts_consensus(parser, sub_program)


import numpy as np

from celescope.tools import utils, analysis_wrapper, plotly_plot
from celescope.tools.tag.analysis_tag import Analysis_tag as At, get_opts_analysis_tag as opts

# default colnames in tsne_tag file 
DEFAULT_COLS = ['', 'tSNE_1', 'tSNE_2', 'cluster', 'Gene_Counts']

class Analysis_tag(At):
    def run(self):
        colnames = list(self.df_tsne_tag.columns)
        tag_cols = set(colnames) - set(DEFAULT_COLS)
        if len(tag_cols) > 1:
            raise ValueError(f"Error: More than one tag barcode in tag_barcode_fasta: {tag_cols}. Currently, Multiple tag barcodes are not supported to display in HTML report.")
        tag_col = list(tag_cols)[0]

        log_tag_col = tag_col + '_log2'
        self.df_tsne_tag[log_tag_col] = np.log2(self.df_tsne_tag[tag_col] + 1)

        report_runner = analysis_wrapper.Report_runner(self.args, display_title=self.display_title)
        df_tsne, _df_marker = report_runner.get_df()

        tsne_cluster = plotly_plot.Tsne_plot(df_tsne, 'cluster').get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)

        tsne_tag = plotly_plot.Tsne_plot(self.df_tsne_tag, log_tag_col, discrete=False).get_plotly_div()
        self.add_data(tsne_tag=tsne_tag)
    
@utils.add_log
def analysis_tag(args):
    with Analysis_tag(args, display_title="Analysis") as runner:
        runner.run()

def get_opts_analysis_tag(parser, sub_program):
    opts(parser, sub_program)


from celescope.flv_trust4.__init__ import REF_DIR
from concurrent.futures import ProcessPoolExecutor
from celescope.tools import utils
import subprocess
import pysam


INDEX = {
    "TCR": ["bcrtcr", "TRA", "TRB"],
    "BCR": ["bcrtcr", "IGH", "IGK", "IGL"],
}


class Mapping_vdj:
    """
    ##Features
    - Extract reads of mapping to VDJ gene.
    
    For TCR library:
    Total Reads: 10,000
    Reads Mapped To Any V(D)J Genes: 9,000
    Reads Mapped To TRA: 4,000
    Reads Mapped To TRB: 5,000
    For BCR library:
    Total Reads: 20,000
    Reads Mapped To Any V(D)J Genes: 15,000
    Reads Mapped To IGH: 5,000
    Reads Mapped To IGK: 5,000
    Reads Mapped To IGL: 5,000
    
    ## Output
    `bcrtcr_1.fq` barcode and umi of mapping to any V(D)J genes reads.
    `bcrtcr_2.fq` sequence of mapping to any V(D)J genes reads.
    """

    def __init__(self, args):
        self.ref = args.ref
        self.fq2 = args.fq2
        self.seqtype = args.seqtype
        self.out_dir = args.out_dir
        self.out_fq1 = f"{self.out_dir}/bcrtcr_1.fq"

        
    def __call__(self):
        utils.check_mkdir(self.out_dir)
        
        results = []
        with ProcessPoolExecutor(max_workers=4) as executor:
            for result in executor.map(self.get_mapping_reads, INDEX[self.seqtype]):
                results.append(result)

        self.write_metrics()
        self.write_fq1()

    @utils.add_log
    def get_mapping_reads(self, index):
        
        cmd = (
            f'fastq-extractor -t 1 '
            f'-f {REF_DIR}/{self.ref}/{index}.fa '
            f'-o {self.out_dir}/{index}_2 '
            f'-u {self.fq2} '
            )
        Mapping_vdj.get_mapping_reads.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def write_metrics(self):
        metrics = []
        input_files = [self.fq2] + [f"{self.out_dir}/{i}_2.fq" for i in INDEX[self.seqtype]]
        
        with ProcessPoolExecutor(max_workers=4) as executor:
            for result in executor.map(utils.get_fastx_read_number, input_files):
                metrics.append(result)

        if self.seqtype == "TCR":
            metrics = dict(zip(["Total Reads" ,"Reads Mapped To Any V(D)J Genes", "Reads Mapped To TRA", "Reads Mapped To TRB"], metrics))
        else:
            metrics = dict(zip(["Total Reads" ,"Reads Mapped To Any V(D)J Genes", "Reads Mapped To IGH", "Reads Mapped To IGK", "Reads Mapped To IGL"], metrics))

        utils.dump_dict_to_json(metrics, f"{self.out_dir}/mapping_VDJ.json")

    @utils.add_log
    def write_fq1(self):
        out_fq1 = open(self.out_fq1, 'w')
        with pysam.FastxFile(f"{self.out_dir}/bcrtcr_2.fq") as f:
            for read in f:
                bc, umi, _ = read.name.split(':')
                new_seq = bc + umi
                new_qual = 'F' * len(new_seq)
                out_fq1.write(utils.fastq_line(read.name, new_seq, new_qual))
        out_fq1.close()


def mapping_vdj(args):
    runner = Mapping_vdj(args)
    runner()


def get_opts_mapping_vdj(parser, sub_program=True):
    if sub_program:
        parser.add_argument("--fq2", help="raw fastq1 file path", required=True)
        parser.add_argument("--ref", help="reference name: hg38 or GRCm38", choices=["hg38", "GRCm38"], required=True)
        parser.add_argument("--seqtype", help="TCR or BCR", choices=["TCR", "BCR"], required=True)
        parser.add_argument("--out_dir", help="out directory", default="mapping_vdj")
    return parser


from celescope.capture_virus.__init__ import __ASSAY__
from celescope.tools.__init__ import STAR_BAM_SUFFIX
from celescope.tools.multi import Multi


class Multi_capture_virus(Multi):
    """
    ## Usage
    1. Make a virus genomeDir using `celescope capture_virus mkref`

    2. Generate shell scripts
    ```
    multi_capture_virus \\
    --mapfile {mapfile} \\
    --virus_genomeDir {virus_genomeDir} \\
    --not_consensus \\
    --thread 4 \\
    --mod shell
    ```

    3. Run shell scripts
    ```
    sh ./shell/{sample}.sh
    ```
    """

    def star_virus(self, sample):
        step = 'star_virus'
        cmd_line = self.get_cmd_line(step, sample)
        if self.args.not_consensus:
            fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        else:
            fq = f'{self.outdir_dic[sample]["consensus"]}/{sample}_consensus.fq'
            cmd_line += ' --consensus_fq '
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def count_virus(self, sample):
        step = 'count_virus'
        cmd_line = self.get_cmd_line(step, sample)
        capture_bam = f'{self.outdir_dic[sample]["star_virus"]}/{sample}_virus_{STAR_BAM_SUFFIX}'
        cmd = (
            f'{cmd_line} '
            f'--capture_bam {capture_bam} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)

    def filter_virus(self, sample):
        step = 'filter_virus'
        cmd_line = self.get_cmd_line(step, sample)
        raw_read_count_file = f'{self.outdir_dic[sample]["count_virus"]}/{sample}_raw_read_count.json'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--raw_read_count_file {raw_read_count_file} '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)

    def analysis_virus(self, sample):
        step = 'analysis_virus'
        cmd_line = self.get_cmd_line(step, sample)
        filter_umi_file = f'{self.outdir_dic[sample]["filter_virus"]}/{sample}_filtered_UMI.csv'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--filter_umi_file {filter_umi_file} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def featureCounts(self, sample):
        step = 'featureCounts'
        cmd_line = self.get_cmd_line(step, sample)
        filter_umi_file = f'{self.outdir_dic[sample]["filter_virus"]}/{sample}_filtered_UMI.csv'
        filter_read_count_json = f'{self.outdir_dic[sample]["filter_virus"]}/{sample}_filtered_read_count.json'
        bam = f'{self.outdir_dic[sample]["star_virus"]}/{sample}_virus_{STAR_BAM_SUFFIX}'
        cmd = (
            f'{cmd_line} '
            f'--filter_umi_file {filter_umi_file} '
            f'--filter_read_count_json {filter_read_count_json} '
            f'--bam {bam}'
        )
        self.process_cmd(cmd, step, sample, m=5, x=self.args.thread)

    def count(self, sample):
        step = 'count'
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_filter_name_sorted.bam'
        filter_umi_file = f'{self.outdir_dic[sample]["filter_virus"]}/{sample}_filtered_UMI.csv'
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--filter_umi_file {filter_umi_file} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=self.args.thread)


def main():
    multi = Multi_capture_virus(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()
